/* And rename to avoid symbol clashes with UXA */
#define fbPolyArc sfbPolyArc
#define fbBlt sfbBlt
#define fbBltOne sfbBltOne
#define fbBltPlane sfbBltPlane
#define fbCopyNtoN sfbCopyNtoN
#define fbCopy1toN sfbCopy1toN
#define fbCopyNto1 sfbCopyNto1
#define fbCopyArea sfbCopyArea
#define fbCopyPlane sfbCopyPlane
#define fbFill sfbFill
#define fbSolidBoxClipped sfbSolidBoxClipped
#define fbPolyFillRect sfbPolyFillRect
#define fbFillSpans sfbFillSpans
#define fbPadPixmap sfbPadPixmap
#define fbValidateGC sfbValidateGC
#define fbGetSpans sfbGetSpans
#define fbPolyGlyphBlt sfbPolyGlyphBlt
#define fbImageGlyphBlt sfbImageGlyphBlt
#define fbPutImage sfbPutImage
#define fbPuXYtImage sfbPutXYImage
#define fbGetImage sfbGetImage
#define fbPolyLine sfbPolyLine
#define fbFixCoordModePrevious sfbFixCoordModePrevious
#define fbPolySegment sfbPolySegment
#define fbBitmapToRegion sfbBitmapToRegion
#define fbPolyPoint sfbPolyPoint
#define fbPushImage sfbPushImage
#define fbPushPixels sfbPushPixels
#define fbSetSpans sfbSetSpans
#define fbSegment sfbSegment
#define fbSegment1 sfbSegment1
#define fbTransparentSpan sfbTransparentSpan
#define fbStipple sfbStipple
#define fbTile sfbTile
#define fbReplicatePixel sfbReplicatePixel

#define fbComposite sfbComposite
#define image_from_pict simage_from_pict
#define free_pixmap_pict sfree_pixmap_pict
